// IKEA Table Lamp Into a Light Alarm With Dimmer by wuwuer3 = Kamil G = instagram.com/wuwuer3
// Published on Instructables on 2020-06-22
// https://www.instructables.com/id/IKEA-Table-Lamp-Into-a-Light-Alarm-With-Dimmer/
// LICENSE: CC-BY-SA
// Use/implement at your own resposibility. Have Fun.

#include <EEPROM.h>
#include "RTClib.h"
RTC_DS1307 rtc;

#include <Adafruit_SSD1306.h>
#define OLED_RESET 6
Adafruit_SSD1306 display(OLED_RESET);

int dimPin = 3; // analog output to an executive MOSFET
int ylwpin = 7; //encoder output; pulled DOWN by an external resistor
int grnpin = 8; //encoder output; pulled DOWN by an external resistor
int clickpin = 4; // button pin - here using internal pullup; see setup()

char* modelabel[6] = {"standby", "details", "set ALARM", "set RTC", "DARK",  "ALARM!"}; // 3 of these labels are used (being actually printed), others are just for your information about the mode meaning
char* alarmsetmask[2] = {"OFF", "ON"};
char* alarmmodemask[3] = {"SMOOTH", "INSTANT", "FLASH"}; // to show them in settings and for your information
char* beepmodemask[4] = {"BEEP OFF", "BEEP 0 min", "BEEP 30 min", "BEEP 60 min"}; // a future feature, not used yet
char* rtctime[] = {};
int dimmask[26] = {0, 1, 2, 3, 4, 8, 12, 20, 26, 33, 41, 50, 59, 69, 80, 92, 104, 118, 132, 147, 163, 180, 197, 216, 235, 255};

// char* colour[2] = {"WHITE", "BLACK"}; // a future feature, not used yet; left here to not be forgotten (it would be nice to have a variable currently activated in SET modes printed as inverted ("BLACK") text)

int tm[6]; // yyyy-mm-dd-hh-mm-ss; an array for received time data
int tm_set[6]; // yyyy-mm-dd-hh-mm-ss; an array for time data to be sent to RTC
int tm_temp[10]; // yyyy-mm-dd-hh-mm-ss-dim-onoff-wakeupmode-beepmode // BEEPMODE is a future feature, not used yet
int tm_alarm[11] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; // yyyy-mm-dd-hh-mm-ss-dim-onoff-wakeupmode-beepmode-status // BEEPMODE is a future feature, not used yet (beeper!)
int temp = 0; // an int to store some temporary values in scrollset() function
float temp2 = 0; // a float to store some temporary values in wakeup() function

int tmrvalue[4]; // an array to store pseudo-random values and their derivatives which are used in turnitoff()
int tm_dif[3] = {0, 0, 0}; // an array to store "time difference" - the dif. between alarm time and current time

int smoothdim; //used in "smooth" wake-up program

const int SMOOTHNESS_CONSTANT = 10; // the number of MINUTES needed to reach the preset DIM value during "smooth" wake-up program; configure it as you wish!!!
const int FLASH_PERIOD = 10; // the number of SECONDS for a full ON-OFF cycle during "flash" wake-up program; configure it as you wish!!!
const int SOLVE_REPETITIONS = 4; // a number of correct(!) repetitions needed to turn off the alarm; configure it as you wish!!!
const bool DIAGNOSTIC_DETAILS = false; // to show or hide more sophisticated diagnostic details: 1) brightness as [n]/255 & 2)the recognized length of your click (the latter is quite useful); configure it as you wish!!!

int tm_mask_up[10] = {2100, 12, 31, 23, 59, 59, 25, 1, 2, 3}; //yyyy-mm-dd-hh-mm-ss-dim-onoff-wakeupmode-beepmode
int tm_mask_dn[10] = {2000, 1, 1, 0, 0, 0, 0, 0, 0, 0}; //yyyy-mm-dd-hh-mm-ss-dim-onoff-wakeupmode-beepmode

bool is_set = false; // used whenever scrollset() is called

// some variables used in time-management in given functions
unsigned long scrollmillis = 0;
unsigned long scrollmillis2 = 0;
unsigned long buttonmillis1 = 0;
unsigned long buttonmillis2 = 0;
unsigned long tickmillis = 0;
unsigned long refreshmillis = 0;

int bms = 0; // button milliseconds - the duration of your last click; may be printed for diagnostic purposes
int bms_old = 0; // button milliseconds - the duration of your last click; same as above

int mode = 0;
int md = 0;
int oldmode = 0;

int dim = 0;
int dim_old = 0;
int dim2 = 0;
int dimstep = 0;

int old_i = 0;

int thr0 = 100; // thresholds used in button() and scrollset()
int thr1 = 20;
int thr2 = 500;
int thr3 = 5000;


void setup() {

  pinMode(LED_BUILTIN, OUTPUT);

  pinMode(clickpin, INPUT_PULLUP);
  pinMode(ylwpin, INPUT);
  pinMode(grnpin, INPUT);
  pinMode(dimPin, OUTPUT);
  analogWrite(dimPin, 0);

  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  display.setTextColor(WHITE);
  display.clearDisplay();
  display.display();

  rtc.begin();
  Serial.begin(9600); //diagnostic purposes (RTC accuracy check), may/should be turned off in normal use!

  gettime();

  //  set the freakin' alarm one minute ahead because of debugging reasons
  //  tm_alarm[3] = tm[3];
  //  tm_alarm[4] = tm[4]+1;
  //  tm_alarm[6] = 25; //full brightness
  //  tm_alarm[7] = 1; // "smooth"
  //  tm_alarm[8] = 0; // ON
  //  set the freakin' alarm one minute ahead because of debugging reasons

  if (EEPROM.read(0) == 255 ) { // "empty" (better: "erased") EEPROM should contain a series of "255" values; this block is to avoid reading bullshit at the first startup after flash and to set the alarm inactive with brightness 0 at 00:00 (together with corresponding eeprom locations)
    for (int i = 0; i < 9; i++) {
      tm_alarm[i] = 0;
    }
    alarm_eepromsave();
  }
  else { // if there's anything better than 255s
    alarm_eepromread();
  }
}

void loop() {

  md = button();

  switch (md) {
    case 1: //short
      if (oldmode == 1) { //from "details" goes to "standby"
        mode = 0;
      }
      else { //from "standby" goes to "details"
        mode = md;
      }
      break;
    case 2: //long
      if (oldmode == 0) { //long click in "standby" -> oled goes dark
        mode = 4;
      }
      else { //long click in "details" -> goes to "set ALARM"
        mode = md;
        alarmset();
      }
      break;
    case 3: //very long click -> goes to "set RTC"
      mode = md;
      rtcset();
      break;
  }

  oldmode = mode;

  dim = scrollset(dim);
  dim = dimcheck(dim);

  if (((millis() - tickmillis) > 200) || (dim_old != dim) || (bms_old != bms)) { // refresh everything possible if anything has happened; this MIGHT be optimized - getting new time is not necessary if you have only changed the dim by one step :)
    gettime();
    oledprint();

    if (tm_alarm[10] != 1) { // nothing will happed if alarm has already started; you cannot change the brightness when alarm is activated!!! (reasonable)
      dimmer(dim);
    }

    dim_old = dim;
    bms_old = bms;

    tickmillis = millis();
  }

  watchit();
  wakeup();
  turnitoff();

}

void watchit() { // watches for the right time to begin the alarm procedure; this will NOT work if the program has stuck somewhere because this function is being called only from loop()

  if (tm_alarm[7] == 1) {
    if (tm[3] == tm_alarm[3] && tm[4] == tm_alarm[4]) {
      tm_alarm[10] = 1;
      mode = 6;
    }
  }

}

void wakeup() { // checks the delay (tm_dif), manages the light according to the chosen "wakeupmode"

  if (tm_alarm[10] == 1) {

    for (int i = 3; i < 6; i++) {
      tm_dif[i - 3] = tm[i] - tm_alarm[i];
    }

    //digitalWrite(LED_BUILTIN, tm[5] % 2); //to flash the Pin13 LED in 1-second interval - use it if you wish

    if (tm_alarm[8] == 0 ) { // "smooth"
      temp2 = float(tm_dif[0]) * 360.0; // hours into seconds
      temp2 += float(tm_dif[1]) * 60.0; // minutes into seconds
      temp2 += float(tm_dif[2]); // seconds plus seconds
      temp2 /= float(SMOOTHNESS_CONSTANT); // divided by CONSTANT (minutes)
      temp2 /= 60.0; // divided by 60 more (due to CONSTANT being given in minutes)

      //temp2 = float(tm_dif[1] * 60.0 + tm_dif[2] / SMOOTHNESS_CONSTANT / 60.0);
      if (temp2 < 1.0) {
        smoothdim = temp2 * float(tm_alarm[6]);
      }
      else {
        smoothdim = tm_alarm[6];
      }
      dimmer(smoothdim);
    }

    if (tm_alarm[8] == 1) { // "instant"
      dimmer(tm_alarm[6]);
    }

    if (tm_alarm[8] == 2) { // "flash"
      temp = float(tm_dif[2] / (0.5 * FLASH_PERIOD)); // (type conversion)
      if (temp % 2 == 0) {
        dimmer(tm_alarm[6]);
      }
      else {
        dimmer(0);
      }
    }

  }

}


void turnitoff () { // to turn off the alarm by engaging you brain; you are asked to give a correct answer for a MODULO division of larger number (0:100) by a smaller number (0:25)

  if (tm_alarm[10] == 1) {

    int i = 0;
    while (i < SOLVE_REPETITIONS) {

      tmrvalue[1] = 0;

      while (tmrvalue[1] == 0) { // this function generates two pseudo-random values for your morning maths; loops until the divisor is NOT equal to zero! :)
        tmrvalue[0] = millis() % 100;
        tmrvalue[1] = millis() % 25;
        tmrvalue[2] = tmrvalue[0] % tmrvalue[1];
        tmrvalue[3] = tmrvalue[0] % 13; // this gives another pseudo-random number to not start from a boring ZERO as a prompt for your answer
      }

      is_set = false;

      oledprint();
      refreshmillis = millis();

      while (is_set == false) {
        temp = tmrvalue[3];
        tmrvalue[3] = scrollset(tmrvalue[3]);
        tmrvalue[3] = tm_check(tmrvalue[3], 6);  //keep it in range 0:25
        if ((tmrvalue[3] != temp) || ((millis() - refreshmillis) > 100)) {
          gettime();
          oledprint();
          wakeup();
          refreshmillis = millis();
        }
      }

      if (is_set == true) {
        if (tmrvalue[3] == tmrvalue[2]) {
          i += 1;
        }
      }
    }

    // after (SOLVE_REPETITIONS) repetitions, the alarm is turned off:
    tm_alarm[10] = 0;
    dimmer(dim);
    mode = 0;
    oledprint();
  }

}

int dimcheck (int dim) {

  if (dim > 25) {
    dim = 25;
  }
  else if (dim < 0) {
    dim = 0;
  }
  return dim;

}

int dimmer(int dim) {

  //dim2 = map(dim, 0, 25, 0, 255); // LINEAR, bad!
  //dim2 = (dim / 25.0) * (dim / 25.0) * 255; // POWER SERIES, still suboptimal!
  dim2 = dimmask[dim];  // MANUALLY created, adjusted POWER-like array; see declaration at the top and MODIFY as you wish!

  analogWrite(dimPin, dim2);

}

int button() {

  if (digitalRead(clickpin) == LOW) {
    buttonmillis1 = millis();
    while (digitalRead(clickpin) == LOW);
    buttonmillis2 = millis() - buttonmillis1;
    bms = buttonmillis2;
    if (buttonmillis2 > thr3) { //very long click
      return 3;
    }
    else if (buttonmillis2 > thr2) { //long click
      return 2;
    }
    else if (buttonmillis2 > thr1) { //short click
      return 1;
    }
    else {  //too short click (bounce!)
      //return 0;
    }
  }
  else {
    return 4; // nothing happened - return 4 is left here for debugging reasons
  }

}

void rtcset() {

  gettime();

  for (int j = 0; j < 6; j++) {
    tm_set[j] = tm[j];
  }
  oledprint();

  for (int j = 0; j < 6; j++) {
    is_set = false;
    refreshmillis = millis();
    while (is_set == false) {
      temp = tm_set[j];
      tm_set[j] = scrollset(tm_set[j]);
      tm_set[j] = tm_check(tm_set[j], j);  // making sure to NOT over-/undervalue month/day/hour/min/sec
      if (tm_set[j] != temp || (millis() - refreshmillis) > 100) {
        gettime();
        oledprint();
        refreshmillis = millis();
      }
    }
  }
  rtc.adjust(DateTime(tm_set[0], tm_set[1], tm_set[2], tm_set[3], tm_set[4], tm_set[5]));
  mode = 1;

}

void alarmset() {

  //serialdebug(); // RTC (in)accuracy measurement; use only if necessary

  for (int j = 0; j < 9; j++) {
    tm_temp[j] = tm_alarm[j];
  }
  oledprint();

  for (int j = 3; j < 9; j++) { //temporarily WITHOUT beepmode at all
    is_set = false;
    if (j == 5) continue;
    refreshmillis = millis();
    while (is_set == false) {
      temp = tm_temp[j];
      tm_temp[j] = scrollset(tm_temp[j]);
      tm_temp[j] = tm_check(tm_temp[j], j);  //make sure to NOT over-/undervalue month/day/hour/min/sec/dim/alarmmode
      if (tm_temp[j] != temp || (millis() - refreshmillis) > 100) {
        gettime();
        oledprint();
        refreshmillis = millis();
      }
    }
  }
  for (int j = 3; j < 9; j++) {
    tm_alarm[j] = tm_temp[j];
  }
  alarm_eepromsave();

  mode = 1;

}

void alarm_eepromsave() { //saves the alarm setting into EEPROM

  for (int j = 3; j < 9; j++) {
    EEPROM.update(j - 3, byte(tm_alarm[j]));
  }

}

void alarm_eepromread() { //gets the last saved alarm setting from EEPROM

  for (int j = 3; j < 9; j++) {
    tm_alarm[j] = EEPROM.read(j - 3);
  }
}

int tm_check(int forw_check, int j) { // makes sure the value does not exceed proper range + loops it at the critical values

  if (forw_check < tm_mask_dn[j]) {
    forw_check = tm_mask_up[j];
  }
  if (forw_check > tm_mask_up[j]) {
    forw_check = tm_mask_dn[j];
  }
  return forw_check;

}

int scrollset(int forw_value) { // accepts a value and allows it to be changed by the movement of rotary encoder
  if (digitalRead(grnpin) != digitalRead(ylwpin)) {
    scrollmillis2 = millis();
    while ((digitalRead(grnpin) == 0) && (digitalRead(clickpin) != LOW)) {
      if (digitalRead(ylwpin) == 0) {
        forw_value = scroll_it(forw_value, 1); // delegates the change to an executive function
        break;
      }
      if (scrollmillis2 - millis() > (5 * thr0)) { // breaks the while() to avoid locking the program when encoder stops in a strange position (my scroll does it every other time)
        break;
      }
    }
  }

  if (digitalRead(grnpin) != digitalRead(ylwpin)) {
    scrollmillis2 = millis();
    while ((digitalRead(ylwpin) == 0) && (digitalRead(clickpin) != LOW)) {
      if (digitalRead(grnpin) == 0) {
        forw_value = scroll_it(forw_value, -1);
        break;
      }
      if (scrollmillis2 - millis() > (5 * thr0)) {
        break;
      }
    }
  }

  if (button() == 1) {
    is_set = true;
  }
  return forw_value;

}

int scroll_it (int scroll_value, int i) { // an executive function which changes the value forwarded by scrollset() in a direction forwarded by scrollset()

  if (((millis() - scrollmillis) < thr0) && i == old_i) { // this is an important part, responsible for a quicker increment/decrement if the scroll keeps going
    scroll_value += (4 * i);
  }
  else {
    scroll_value += i;
  }
  old_i = i; //direction
  scrollmillis = millis();
  return scroll_value;

}

void gettime () { // simply gets time from RTC module

  DateTime now = rtc.now();
  tm[0] = now.year();
  tm[1] = now.month();
  tm[2] = now.day();
  tm[3] = now.hour();
  tm[4] = now.minute();
  tm[5] = now.second();

}

void oledprint() { // prints every necessary part, depending on the "mode" value; may be (thus is) called at any time!

  display.clearDisplay();

  //ALARM!
  if (mode == 6) {
    display.setCursor(0, 0);
    display.setTextSize(2);

    display.print(tm[3] / 10); // division and modulo division is a trick to obtain leading zeros in values below 10
    display.print(tm[3] % 10); // division and modulo division is a trick to obtain leading zeros in values below 10
    display.print(":");
    display.print(tm[4] / 10);
    display.print(tm[4] % 10);
    display.print(":");
    display.print(tm[5] / 10);
    display.print(tm[5] % 10);

    display.setCursor(0, 16);
    //display.setTextSize(1);
    display.print(tmrvalue[0]);
    display.print("%");
    display.print(tmrvalue[1]);
    display.print("=? ");
    display.print(tmrvalue[3]);

    display.display();
    return; // the work has been done, now skipping the rest of IFs
  }


  //dark
  // a single dot is moving in x (minutes) and y (hours) axes - just to signal that anything is working; you can comment it out
  if (mode == 4) {
    int x = map(tm[4], 0, 60, 0, 127);
    int y = map(tm[3], 0, 23, 0, 31);
    display.drawPixel(x, y, WHITE);
    display.display();
    return; // the work has been done, now skipping the rest of IFs
  }

  //standby
  //standard, "clock-like" wiew
  if (mode ==  0) {
    display.setCursor(0, 2);
    display.setTextSize(4);
    display.print(tm[3] / 10);
    display.print(tm[3] % 10);
    display.setCursor(69, 2);
    display.print(tm[4] / 10);
    display.print(tm[4] % 10);
    display.setTextSize(1);
    display.setCursor(54, 8);
    display.print(tm[5] / 10);
    display.setCursor(54, 17);
    display.print(tm[5] % 10);
    if (tm_alarm[7] == 1) { // a hash is printed if alarm is set ON
      display.setCursor(120, 8);
      display.print("#");
    }
    if (tm_alarm[10] == 1) { // a hash is printed if alarm has started - this feature is no longer visible and necessary
      display.setCursor(120, 16);
      display.print("#");
    }
  }


  //details, printed in details/alarmset/rtcset mode
  if (mode != 0 && mode != 4) {
    display.setCursor(0, 0);
    display.setTextSize(1);

    if (DIAGNOSTIC_DETAILS == true && mode == 1) {
      display.print("DIAG:");
    }
    else {
      display.print(modelabel[mode]); //MODE
    }

    if (DIAGNOSTIC_DETAILS == true) {
      display.setCursor(36, 0);
      display.print(dim2);
      display.print("/255");
      display.print(".");
      display.print(bms);
      display.print("ms ");
    }
    else {
      display.setCursor(60, 0);
      display.print(dim);
      display.print("*");
    }

    display.setCursor(0, 8);
    display.print(tm[0]);
    display.print("-");
    display.print(tm[1] / 10);
    display.print(tm[1] % 10);
    display.print("-");
    display.print(tm[2] / 10);
    display.print(tm[2] % 10);
    display.print(" ");
    display.print(tm[3] / 10);
    display.print(tm[3] % 10);
    display.print(":");
    display.print(tm[4] / 10);
    display.print(tm[4] % 10);
    display.print(":");
    display.print(tm[5] / 10);
    display.print(tm[5] % 10);
    if (tm_alarm[7] == 1) {
      display.setCursor(120, 8);
      display.print("#");
    }
    if (tm_alarm[10] == 1) {
      display.setCursor(120, 16);
      display.print("#");
    }
  }

  //alarmset
  if (mode == 2) {
    display.setCursor(0, 16);
    display.setTextSize(1);
    display.print(tm_temp[3] / 10);
    display.print(tm_temp[3] % 10);
    display.print(":");
    display.print(tm_temp[4] / 10);
    display.print(tm_temp[4] % 10);
    display.print(" ");
    display.print(tm_temp[6]);
    display.print("*");
    display.setCursor(60, 16);
    display.print(alarmsetmask[tm_temp[7]]);
    display.setCursor(84, 16);
    display.print(alarmmodemask[tm_temp[8]]);
  }

  //rtcset
  if (mode == 3) {
    display.setCursor(0, 24);
    display.setTextSize(1);
    display.print(tm_set[0]);
    display.print("-");
    display.print(tm_set[1] / 10);
    display.print(tm_set[1] % 10);
    display.print("-");
    display.print(tm_set[2] / 10);
    display.print(tm_set[2] % 10);
    display.print(" ");
    display.print(tm_set[3] / 10);
    display.print(tm_set[3] % 10);
    display.print(":");
    display.print(tm_set[4] / 10);
    display.print(tm_set[4] % 10);
    display.print(":");
    display.print(tm_set[5] / 10);
    display.print(tm_set[5] % 10);
  }

  display.display();

}

void serialdebug() { // a function for RTC (in)accuracy measurement; called from alarmset() if necessary

  gettime();

  Serial.print(tm[0]);
  Serial.print("-");
  Serial.print(tm[1]);
  Serial.print("-");
  Serial.print(tm[2]);
  Serial.print(" ");
  Serial.print(tm[3]);
  Serial.print(":");
  Serial.print(tm[4]);
  Serial.print(":");
  Serial.println(tm[5]);

}
